/*
 *  <Short Description>
 *  Copyright (C) 2024  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <blt/gfx/renderer/font_renderer.h>

namespace blt::gfx
{
    font_texture_atlas::font_texture_atlas(blt::i32 dimensions): texture_gl2D(dimensions, dimensions, GL_RED)
    {}
    
    blt::u64 font_texture_atlas::add_font(const font::font_file_t& file)
    {
        return add_font(file, file.character_min_index);
    }
    
    blt::u64 font_texture_atlas::add_font(const font::font_file_t& file, blt::u64 start)
    {
        blt::u32 max_height = 0;
        for (auto i = start; i < file.character_max_index; i++)
        {
            // returns an empty optional if there is no error code
            if (file.font.load_char(i))
                continue;
            auto face = *file.font.get();
            
            auto width = face->glyph->bitmap.width;
            auto height = face->glyph->bitmap.rows;
            max_height = std::max(max_height, height);
            
            // if adding this width overflows, we need to move to the next row.
            if (used_width + width > static_cast<blt::u32>(getWidth()))
            {
                used_height += max_height;
                max_height = 0;
                used_width = 0;
            }
            
            // if we can't fit the height of this character we should move to the next texture atlas.
            if (used_height + height > static_cast<blt::u32>(getHeight()))
                return i;
            
            auto begin_width = used_width;
            auto begin_height = used_height;
            auto end_width = used_width + width;
            auto end_height = used_height + height;
            
            // upload the texture
            upload(face->glyph->bitmap.buffer, GL_RED, 0, static_cast<blt::i32>(begin_width), static_cast<blt::i32>(begin_height),
                   static_cast<blt::i32>(width), static_cast<blt::i32>(height));
            
            glyphs.insert(std::pair{i, bounded_t{
                    font::glyph_t{
                            {width, height},
                            {face->glyph->bitmap_left, face->glyph->bitmap_top},
                            face->glyph->advance.x},
                    {begin_width, begin_height}, {end_width, end_height}}
            });
        }
        return -1;
    }
    
    font_renderer_t::font_renderer_t(const std::vector<font::font_file_t>& files, blt::i32 dimensions)
    {
    
    }
}